import typing

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_control, cache_page
from django.views.generic import FormView, RedirectView, TemplateView
from terminusgps.mixins import HtmxTemplateResponseMixin
from terminusgps_payments.models import (
    AddressProfile,
    CustomerProfile,
    PaymentProfile,
)

from terminusgps_notifications.forms import CustomerSubscriptionCreationForm
from terminusgps_notifications.models import Customer, WialonToken


@method_decorator(cache_page(timeout=60 * 15), name="dispatch")
@method_decorator(cache_control(private=True), name="dispatch")
class DashboardView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, TemplateView
):
    content_type = "text/html"
    extra_context = {"title": "Dashboard"}
    http_method_names = ["get"]
    partial_template_name = (
        "terminusgps_notifications/customers/partials/_dashboard.html"
    )
    template_name = "terminusgps_notifications/customers/dashboard.html"

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        context["customer"] = Customer.objects.get(user=self.request.user)
        return context


class AccountView(LoginRequiredMixin, HtmxTemplateResponseMixin, TemplateView):
    content_type = "text/html"
    extra_context = {"title": "Account"}
    http_method_names = ["get"]
    partial_template_name = (
        "terminusgps_notifications/customers/partials/_account.html"
    )
    template_name = "terminusgps_notifications/customers/account.html"

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        try:
            customer, _ = Customer.objects.get_or_create(
                user=self.request.user
            )
            customer_profile = CustomerProfile.objects.get(
                user=self.request.user
            )
            context["customer"] = customer
            context["customer_profile"] = customer_profile
        except CustomerProfile.DoesNotExist:
            context["customer"] = customer
            context["customer_profile"] = None
        return context


class SubscriptionView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, TemplateView
):
    content_type = "text/html"
    extra_context = {"title": "Subscription"}
    http_method_names = ["get"]
    partial_template_name = (
        "terminusgps_notifications/customers/partials/_subscription.html"
    )
    template_name = "terminusgps_notifications/customers/subscription.html"


class NotificationsView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, TemplateView
):
    content_type = "text/html"
    extra_context = {"title": "Notifications"}
    http_method_names = ["get"]
    partial_template_name = (
        "terminusgps_notifications/customers/partials/_notifications.html"
    )
    template_name = "terminusgps_notifications/customers/notifications.html"

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if name := request.GET.get("access_token"):
            token = WialonToken(name=name)
            try:
                customer = Customer.objects.get(
                    user__username=request.GET.get("user", "")
                )
                if not hasattr(customer, "token"):
                    token.customer = customer
                    token.save()
                    messages.add_message(
                        request, 10, "Wialon API token was saved successfully!"
                    )
            except Customer.DoesNotExist:
                return HttpResponse(status=404)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        try:
            context["customer"] = Customer.objects.get(user=self.request.user)
            context["has_token"] = hasattr(context["customer"], "token")
        except Customer.DoesNotExist:
            context["customer"] = None
            context["has_token"] = None
        return context


class CustomerSubscriptionCreateView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, FormView
):
    content_type = "text/html"
    extra_context = {"title": "Create Subscription"}
    form_class = CustomerSubscriptionCreationForm
    http_method_names = ["get", "post"]
    partial_template_name = "terminusgps_notifications/customers/partials/_create_subscription.html"
    template_name = (
        "terminusgps_notifications/customers/create_subscription.html"
    )

    def get_form(self, form_class=None) -> CustomerSubscriptionCreationForm:
        form = super().get_form(form_class=form_class)
        payment_qs = PaymentProfile.objects.for_user(self.request.user)
        address_qs = AddressProfile.objects.for_user(self.request.user)
        form.fields["payment_profile"].queryset = payment_qs
        form.fields["payment_profile"].empty_label = None
        form.fields["address_profile"].queryset = address_qs
        form.fields["address_profile"].empty_label = None
        return form

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        try:
            context["customer"] = Customer.objects.get(user=self.request.user)
        except Customer.DoesNotExist:
            context["customer"] = None
        return context

    def form_valid(
        self, form: CustomerSubscriptionCreationForm
    ) -> HttpResponse:
        return HttpResponseRedirect(
            reverse("terminusgps_notifications:subscription"),
            headers={"HX-Retarget": "#subscription"},
        )


class CustomerStatsView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, TemplateView
):
    content_type = "text/html"
    http_method_names = ["get"]
    partial_template_name = (
        "terminusgps_notifications/customers/partials/_stats.html"
    )
    template_name = "terminusgps_notifications/customers/stats.html"

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        try:
            context["customer"] = Customer.objects.get(user=self.request.user)
        except Customer.DoesNotExist:
            context["customer"] = None
        return context


class WialonLoginView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, RedirectView
):
    http_method_names = ["get"]
    permanent = True
    query_string = True
    url = "https://hosting.terminusgps.com/login.html"


class WialonCallbackView(HtmxTemplateResponseMixin, TemplateView):
    content_type = "text/html"
    extra_context = {"title": "Logged Into Wialon"}
    http_method_names = ["get"]
    partial_template_name = (
        "terminusgps_notifications/customers/partials/_wialon_callback.html"
    )
    template_name = "terminusgps_notifications/customers/wialon_callback.html"

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if name := request.GET.get("access_token"):
            token = WialonToken(name=name)
        else:
            return HttpResponse(status=406)
        try:
            token.customer = Customer.objects.get(
                user__username=request.GET.get("user", "")
            )
            token.save()
            messages.add_message(
                request, 10, "Wialon API token was successfully saved!"
            )
            return super().get(request, *args, **kwargs)
        except Customer.DoesNotExist:
            return HttpResponse(status=404)



import json
import logging
import typing
from urllib.parse import urlencode

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ValidationError
from django.db.models import QuerySet
from django.forms import Form
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import cache_control, cache_page
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    FormView,
    ListView,
    TemplateView,
    UpdateView,
)
from terminusgps.mixins import HtmxTemplateResponseMixin
from terminusgps.wialon.session import WialonAPIError, WialonSession

from terminusgps_notifications import constants, forms, models

logger = logging.getLogger(__name__)


@method_decorator(cache_page(timeout=60 * 15), name="get")
@method_decorator(cache_control(private=True), name="get")
class WialonNotificationUnitSelectFormView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, FormView
):
    content_type = "text/html"
    form_class = forms.WialonUnitSelectionForm
    http_method_names = ["get", "post"]
    partial_template_name = (
        "terminusgps_notifications/units/partials/_select.html"
    )
    template_name = "terminusgps_notifications/units/select.html"

    def get_success_url(self, form) -> str:
        return reverse(
            "terminusgps_notifications:select triggers",
            query={"un": str(form.cleaned_data["units"])},
        )

    def form_valid(self, form):
        return HttpResponseRedirect(self.get_success_url(form))

    def get_form(self, form_class=None) -> forms.WialonUnitSelectionForm:
        form = super().get_form(form_class=form_class)
        customer = models.Customer.objects.get(user=self.request.user)
        if hasattr(customer, "token"):
            try:
                token = getattr(customer, "token").name
                with WialonSession(token=token) as session:
                    unit_list = [
                        (int(unit["id"]), str(unit["nm"]))
                        for unit in customer.get_units_from_wialon(session)
                    ]
                    form.fields["units"].choices = unit_list
            except WialonAPIError as e:
                logger.warning(e)
        return form


class WialonNotificationTriggerSelectFormView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, FormView
):
    content_type = "text/html"
    form_class = forms.TriggerForm
    http_method_names = ["get", "post"]
    initial = {"t": constants.WialonNotificationTriggerType.SENSOR}
    partial_template_name = (
        "terminusgps_notifications/triggers/partials/_select.html"
    )
    template_name = "terminusgps_notifications/triggers/select.html"

    def get_initial(self, **kwargs) -> dict[str, typing.Any]:
        initial: dict[str, typing.Any] = super().get_initial(**kwargs)
        initial["un"] = self.request.GET.get("un")
        return initial

    def form_valid(self, form: forms.TriggerForm) -> HttpResponse:
        return HttpResponseRedirect(
            reverse(
                "terminusgps_notifications:create notifications",
                query={
                    "un": form.cleaned_data["un"],
                    "p": json.dumps(form.cleaned_data["p"]),
                    "t": form.cleaned_data["t"],
                },
            )
        )


class WialonNotificationTriggerParametersFormSuccessView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, TemplateView
):
    content_type = "text/html"
    http_method_names = ["get"]
    template_name = (
        "terminusgps_notifications/triggers/parameters_success.html"
    )
    partial_template_name = (
        "terminusgps_notifications/triggers/partials/_parameters_success.html"
    )

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        context["value"] = self.request.GET.get("p")
        return context


class WialonNotificationTriggerParametersFormView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, FormView
):
    content_type = "text/html"
    http_method_names = ["get", "post"]
    partial_template_name = (
        "terminusgps_notifications/triggers/partials/_parameters.html"
    )
    template_name = "terminusgps_notifications/triggers/parameters.html"

    def setup(self, request: HttpRequest, *args, **kwargs) -> None:
        super().setup(request, *args, **kwargs)
        self.trigger = None
        if t := self.request.GET.get("t"):
            self.trigger = t
        elif t := self.request.POST.get("t"):
            self.trigger = t

    def get_form_class(self):
        if self.trigger is None:
            return Form
        return forms.triggers.TRIGGER_FORM_MAP.get(self.trigger)

    def form_valid(self, form):
        return HttpResponseRedirect(
            reverse(
                "terminusgps_notifications:trigger parameters success",
                query={"p": form.as_json()},
            )
        )


class WialonNotificationCreateView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, CreateView
):
    extra_context = {
        "message_tags": constants.WialonNotificationMessageTag.choices
    }
    content_type = "text/html"
    form_class = forms.WialonNotificationCreationForm
    http_method_names = ["get", "post"]
    model = models.WialonNotification
    partial_template_name = (
        "terminusgps_notifications/notifications/partials/_create.html"
    )
    success_url = reverse_lazy("terminusgps_notifications:list notifications")
    template_name = "terminusgps_notifications/notifications/create.html"

    def get_initial(self, **kwargs) -> dict[str, typing.Any]:
        initial: dict[str, typing.Any] = super().get_initial(**kwargs)
        schedule = {"f1": 0, "f2": 0, "t1": 0, "t2": 0, "m": 0, "w": 0, "y": 0}
        customer = models.Customer.objects.get(user=self.request.user)
        trigger = {
            "t": self.request.GET.get("t", ""),
            "p": json.loads(self.request.GET.get("p", "{}")),
        }

        initial["name"] = "New Notification"
        initial["message"] = "%NOTIFICATION%"
        initial["timezone"] = 0
        initial["customer"] = customer
        initial["unit_list"] = self.request.GET.get("un")
        initial["trigger"] = trigger
        initial["schedule"] = schedule.copy()
        initial["control_schedule"] = schedule.copy()
        return initial

    def form_valid(
        self, form: forms.WialonNotificationCreationForm
    ) -> HttpResponse:
        try:
            customer = form.cleaned_data["customer"]
            token = getattr(customer, "token").name
            with WialonSession(token=token) as session:
                notification = form.save(commit=False)
                notification.actions = notification.get_actions()
                notification.text = notification.get_text()
                api_response = notification.create_in_wialon(session)
                notification.wialon_id = int(api_response[0])
                notification.save()
            return super().form_valid(form=form)
        except WialonAPIError as e:
            form.add_error(
                None,
                ValidationError(
                    _("Whoops! '%(error_message)s'"),
                    code="invalid",
                    params={"error_message": str(e)},
                ),
            )
            return self.form_invalid(form=form)


class WialonNotificationDetailView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, DetailView
):
    content_type = "text/html"
    http_method_names = ["get"]
    model = models.WialonNotification
    partial_template_name = (
        "terminusgps_notifications/notifications/partials/_detail.html"
    )
    pk_url_kwarg = "notification_pk"
    template_name = "terminusgps_notifications/notifications/detail.html"

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        context["title"] = self.get_object().name
        return context

    def get_queryset(self) -> QuerySet:
        return super().get_queryset().filter(customer__user=self.request.user)


class WialonNotificationUpdateView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, UpdateView
):
    extra_context = {
        "message_tags": constants.WialonNotificationMessageTag.choices
    }
    content_type = "text/html"
    http_method_names = ["get", "post"]
    form_class = forms.WialonNotificationUpdateForm
    model = models.WialonNotification
    partial_template_name = (
        "terminusgps_notifications/notifications/partials/_update.html"
    )
    pk_url_kwarg = "notification_pk"
    template_name = "terminusgps_notifications/notifications/update.html"

    def get_queryset(self) -> QuerySet:
        return super().get_queryset().filter(customer__user=self.request.user)

    def form_valid(
        self, form: forms.WialonNotificationUpdateForm
    ) -> HttpResponse:
        notification = form.save(commit=True)
        token = getattr(notification.customer, "token").name
        if form.changed_data:
            try:
                with WialonSession(token=token) as session:
                    if "method" in form.changed_data:
                        notification.actions = notification.get_actions()
                    if "message" in form.changed_data:
                        notification.text = notification.get_text()
                    notification.update_in_wialon(session)
            except (ValueError, WialonAPIError) as e:
                form.add_error(
                    None,
                    ValidationError(
                        _("Whoops! %(error)s"),
                        code="invalid",
                        params={"error": str(e)},
                    ),
                )
                return self.form_invalid(form=form)
        return HttpResponseRedirect(
            notification.get_absolute_url(), headers={"HX-Refresh": "true"}
        )


class WialonNotificationDeleteView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, DeleteView
):
    content_type = "text/html"
    http_method_names = ["get", "post"]
    model = models.WialonNotification
    partial_template_name = (
        "terminusgps_notifications/notifications/partials/_delete.html"
    )
    pk_url_kwarg = "notification_pk"
    template_name = "terminusgps_notifications/notifications/delete.html"
    success_url = reverse_lazy("terminusgps_notifications:list notifications")

    def get_queryset(self) -> QuerySet:
        return super().get_queryset().filter(customer__user=self.request.user)

    def form_valid(self, form: Form) -> HttpResponse:
        try:
            token = getattr(self.object.customer, "token").name
            with WialonSession(token=token) as session:
                self.object.delete_in_wialon(session)
            return super().form_valid(form=form)
        except (ValueError, WialonAPIError):
            return HttpResponse(status=406)


class WialonNotificationListView(
    LoginRequiredMixin, HtmxTemplateResponseMixin, ListView
):
    allow_empty = True
    content_type = "text/html"
    http_method_names = ["get"]
    model = models.WialonNotification
    ordering = "name"
    partial_template_name = (
        "terminusgps_notifications/notifications/partials/_list.html"
    )
    template_name = "terminusgps_notifications/notifications/list.html"

    def get_context_data(self, **kwargs) -> dict[str, typing.Any]:
        context: dict[str, typing.Any] = super().get_context_data(**kwargs)
        customer = models.Customer.objects.get(user=self.request.user)
        context["has_token"] = hasattr(customer, "token")
        context["login_params"] = urlencode(
            {
                "client_id": "Terminus GPS Notifications",
                "access_type": settings.WIALON_TOKEN_ACCESS_TYPE,
                "activation_time": 0,
                "duration": 2_592_000,
                "lang": "en",
                "flags": 0x1,
                "user": self.request.user.username,
                "redirect_uri": "http://127.0.0.1:8000/notifications/",
                "response_type": "token",
            }
        )
        return context

    def get_ordering(self) -> str:
        if user_input := self.request.GET.get("order"):
            if user_input in ("name", "date_created"):
                return user_input
        return self.ordering

    def get_queryset(self) -> QuerySet:
        return (
            super()
            .get_queryset()
            .filter(customer__user=self.request.user)
            .order_by(self.get_ordering())
        )



from django.contrib import admin

from . import models


@admin.register(models.Customer)
class CustomerAdmin(admin.ModelAdmin):
    list_display = ["user", "sms_count", "voice_count", "subscription__status"]
    list_filter = ["subscription__status"]
    readonly_fields = ["tax", "grand_total"]
    exclude = ["subscription"]
    fieldsets = [
        (None, {"fields": ["user", "date_format", "resource_id"]}),
        (
            "Messaging",
            {
                "fields": [
                    "max_sms_count",
                    "max_voice_count",
                    "sms_count",
                    "voice_count",
                ]
            },
        ),
        (
            "Pricing",
            {"fields": ["tax_rate", "subtotal", "tax", "grand_total"]},
        ),
    ]


@admin.register(models.WialonToken)
class WialonTokenAdmin(admin.ModelAdmin):
    list_display = ["customer"]
    exclude = ["name"]
    readonly_fields = ["flags"]


@admin.register(models.WialonNotification)
class WialonNotificationAdmin(admin.ModelAdmin):
    list_display = ["name", "customer"]
    list_filter = ["customer"]
    readonly_fields = [
        "text",
        "wialon_id",
        "customer",
        "actions",
        "schedule",
        "control_schedule",
        "trigger",
        "unit_list",
    ]



from django import forms
from django.utils.translation import gettext_lazy as _


class WialonUnitSelectionForm(forms.Form):
    units = forms.TypedMultipleChoiceField(
        choices=[],
        help_text=_("Ctrl+click to select multiple. Cmd+click on Mac."),
        widget=forms.widgets.SelectMultiple(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 user-invalid:bg-red-50 user-invalid:text-red-600",
                "autofocus": True,
            }
        ),
        coerce=int,
    )



from django.urls import path

from . import views

app_name = "terminusgps_notifications"
urlpatterns = [
    path("", views.HomeView.as_view(), name="home"),
    path("dashboard/", views.DashboardView.as_view(), name="dashboard"),
    path("stats/", views.CustomerStatsView.as_view(), name="stats"),
    path("account/", views.AccountView.as_view(), name="account"),
    path(
        "subscription/", views.SubscriptionView.as_view(), name="subscription"
    ),
    path(
        "subscriptions/create/",
        views.CustomerSubscriptionCreateView.as_view(),
        name="create subscriptions",
    ),
    path(
        "notifications/",
        views.NotificationsView.as_view(),
        name="notifications",
    ),
    path(
        "notifications/list/",
        views.WialonNotificationListView.as_view(),
        name="list notifications",
    ),
    path(
        "notifications/create/",
        views.WialonNotificationCreateView.as_view(),
        name="create notifications",
    ),
    path(
        "notifications/<int:notification_pk>/detail/",
        views.WialonNotificationDetailView.as_view(),
        name="detail notifications",
    ),
    path(
        "notifications/<int:notification_pk>/update/",
        views.WialonNotificationUpdateView.as_view(),
        name="update notifications",
    ),
    path(
        "notifications/<int:notification_pk>/delete/",
        views.WialonNotificationDeleteView.as_view(),
        name="delete notifications",
    ),
    path(
        "notifications/units/select/",
        views.WialonNotificationUnitSelectFormView.as_view(),
        name="select units",
    ),
    path(
        "notifications/triggers/select/",
        views.WialonNotificationTriggerSelectFormView.as_view(),
        name="select triggers",
    ),
    path(
        "notifications/triggers/parameters/",
        views.WialonNotificationTriggerParametersFormView.as_view(),
        name="trigger parameters",
    ),
    path(
        "notifications/triggers/parameters/success/",
        views.WialonNotificationTriggerParametersFormSuccessView.as_view(),
        name="trigger parameters success",
    ),
    path(
        "wialon/login/", views.WialonLoginView.as_view(), name="wialon login"
    ),
    path(
        "wialon/callback/",
        views.WialonCallbackView.as_view(),
        name="wialon callback",
    ),
]



from .notifications import (
    WialonNotificationCreationForm,
    WialonNotificationUpdateForm,
)
from .subscriptions import CustomerSubscriptionCreationForm
from .triggers import *
from .units import WialonUnitSelectionForm



import ast
import logging
import typing
import urllib.parse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.validators import MaxValueValidator, MinLengthValidator
from django.db import models, transaction
from django.db.models import F
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from encrypted_field import EncryptedField
from terminusgps.validators import validate_is_digit
from terminusgps.wialon import flags
from terminusgps.wialon.session import WialonAPIError, WialonSession

logger = logging.getLogger(__name__)


class Customer(models.Model):
    """A Terminus GPS Notifications customer."""

    user = models.OneToOneField(get_user_model(), on_delete=models.CASCADE)
    """Django user."""
    date_format = models.CharField(
        choices=[
            ("%Y-%m-%d %H:%M:%S", "YYYY-MM-DD HH:MM:SS"),
            ("%Y-%m-%d %H:%M", "YYYY-MM-DD HH:MM"),
        ],
        default="%Y-%m-%d %H:%M:%S",
        help_text="Select a date format for notification messages.",
        max_length=24,
    )
    """Date format for notifications."""
    resource_id = models.CharField(
        blank=True,
        default=None,
        help_text="Enter an 8-digit Wialon resource id to create notifications in.",
        max_length=8,
        null=True,
        validators=[validate_is_digit, MinLengthValidator(8)],
    )
    """Wialon resource id."""
    max_sms_count = models.PositiveIntegerField(
        default=500,
        help_text="Enter the maximum number of allowed sms messages in a single period.",
    )
    """Maximum number of allowed sms messages for the period."""
    max_voice_count = models.PositiveIntegerField(
        default=500,
        help_text="Enter the maximum number of allowed voice messages in a single period.",
    )
    """Maximum number of allowed voice messages for the period."""
    sms_count = models.PositiveIntegerField(default=0)
    """Current number of sms messages for the period."""
    voice_count = models.PositiveIntegerField(default=0)
    """Current number of voice messages for the period."""

    tax_rate = models.DecimalField(
        max_digits=9,
        decimal_places=4,
        default=0.0825,
        help_text="Enter a tax rate as a decimal.",
    )
    """Subscription tax rate."""
    subtotal = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        default=44.99,
        help_text="Enter a dollar amount to charge the customer (not incl. tax) every period.",
    )
    """Subscription subtotal."""
    tax = models.GeneratedField(
        expression=(F("subtotal") * (F("tax_rate") + 1)) - F("subtotal"),
        output_field=models.DecimalField(max_digits=9, decimal_places=2),
        db_persist=True,
        help_text="Automatically generated tax amount.",
    )
    """Subscription tax total."""
    grand_total = models.GeneratedField(
        expression=F("subtotal") * (F("tax_rate") + 1),
        output_field=models.DecimalField(max_digits=9, decimal_places=2),
        db_persist=True,
        help_text="Automatically generated grand total amount (subtotal+tax).",
    )
    """Subscription grand total (subtotal + tax)."""

    subscription = models.ForeignKey(
        "terminusgps_payments.Subscription",
        on_delete=models.CASCADE,
        related_name="customer",
        null=True,
        blank=True,
        default=None,
    )
    """Associated subscription."""

    class Meta:
        verbose_name = _("customer")
        verbose_name_plural = _("customers")

    def __str__(self) -> str:
        """Returns the customer user."""
        return str(self.user)

    def get_units_from_wialon(
        self, session: WialonSession, force: bool = False
    ) -> list[dict[str, typing.Any]]:
        """
        Returns a list of of customer Wialon unit dictionaries from the Wialon API.

        Wialon unit dictionary format:

        +------------+---------------+---------------------------+
        | key        | type          | desc                      |
        +============+===============+===========================+
        | ``"mu"``   | :py:obj:`int` | Measurement system        |
        +------------+---------------+---------------------------+
        | ``"nm"``   | :py:obj:`str` | Unit name                 |
        +------------+---------------+---------------------------+
        | ``"cls"``  | :py:obj:`int` | Superclass ID: 'avl_unit' |
        +------------+---------------+---------------------------+
        | ``"id"``   | :py:obj:`int` | Unit ID                   |
        +------------+---------------+---------------------------+
        | ``"uacl"`` | :py:obj:`int` | User's access rights      |
        +------------+---------------+---------------------------+

        :param force: Whether to force a Wialon API call instead of using a cached response. Default is :py:obj:`False` (use cache).
        :type force: bool
        :returns: A list of Wialon unit dictionaries.
        :rtype: list[dict[str, ~typing.Any]]

        """
        return session.wialon_api.core_search_items(
            **{
                "spec": {
                    "itemsType": "avl_unit",
                    "propName": "sys_id",
                    "propValueMask": "*",
                    "sortType": "sys_id",
                    "propType": "property",
                },
                "force": int(force),
                "flags": flags.DataFlag.UNIT_BASE,
                "from": 0,
                "to": 0,
            }
        ).get("items", [])


class WialonToken(models.Model):
    """Wialon API token."""

    customer = models.OneToOneField(
        "terminusgps_notifications.Customer",
        on_delete=models.CASCADE,
        related_name="token",
    )
    """Associated customer."""

    name = EncryptedField(max_length=72)
    """Wialon API token name."""
    flags = models.PositiveIntegerField(
        default=settings.WIALON_TOKEN_ACCESS_TYPE
    )
    """Wialon token flags."""

    def __str__(self) -> str:
        """Returns '<customer email>'s WialonToken'."""
        return f"{self.customer}'s WialonToken"


class WialonNotification(models.Model):
    """Wialon notification."""

    class WialonNotificationMethod(models.TextChoices):
        """Wialon notification method."""

        SMS = "sms", _("SMS")
        VOICE = "voice", _("Voice")

    name = models.CharField(
        max_length=64, help_text="Provide a memorable name."
    )
    """Notification name."""
    message = models.CharField(max_length=1024, help_text="Enter a message.")
    """Notification message."""
    method = models.CharField(
        choices=WialonNotificationMethod.choices,
        default=WialonNotificationMethod.SMS,
        help_text="Select a delivery method.",
        max_length=5,
    )
    """Notification method."""

    activation_time = models.DateTimeField(
        blank=True,
        default=None,
        help_text="Provide a valid date and time in the format: YYYY-MM-DD HH:MM:SS. Leave this blank to activate immediately.",
        null=True,
    )
    """Activation date/time."""
    deactivation_time = models.DateTimeField(
        blank=True,
        default=None,
        help_text="Provide a valid date and time in the format: YYYY-MM-DD HH:MM:SS. Leave this blank to never deactivate.",
        null=True,
    )
    """Deactivation date/time."""
    max_alarms = models.PositiveIntegerField(
        default=0,
        help_text="Provide the maximum number of alarms. 0 = unlimited alarms.",
    )
    """Maximum number of alarms (0 = unlimited)."""
    max_message_interval = models.PositiveIntegerField(
        default=3600,
        choices=[
            (0, _("Any time")),
            (60, _("1 minute")),
            (600, _("10 minutes")),
            (1800, _("30 minutes")),
            (3600, _("1 hour")),
            (21600, _("6 hours")),
            (43200, _("12 hours")),
            (86400, _("1 day")),
            (864000, _("10 days")),
        ],
        help_text="Select the maximum allowed time between messages.",
    )
    """Max time interval between messages."""
    alarm_timeout = models.PositiveIntegerField(
        default=0,
        validators=[MaxValueValidator(1800)],
        help_text="Provide the alarm timeout in seconds. 0 = never timeout.",
    )
    """Alarm timeout in seconds. Max is ``1800`` (30 minutes in seconds)."""
    control_period = models.PositiveIntegerField(
        default=3600,
        choices=[
            (0, _("Any time")),
            (60, _("Last minute")),
            (600, _("Last 10 minutes")),
            (3600, _("Last hour")),
            (86400, _("Last day")),
        ],
        help_text="Select a control period relative to current time.",
    )
    """Control period relative to current time in seconds."""
    min_duration_alarm = models.PositiveIntegerField(
        default=0,
        validators=[MaxValueValidator(86400)],
        help_text="Provide the minimum duration of alarm state in seconds.",
    )
    """Minimum duration of alarm state in seconds. Max is ``86400`` (1 day)."""
    min_duration_prev = models.PositiveIntegerField(
        default=0, validators=[MaxValueValidator(86400)]
    )
    """Minimum duration of previous state in seconds. Max is ``86400`` (1 day)."""
    language = models.CharField(
        max_length=2,
        default="en",
        choices=[("en", _("English"))],
        help_text="Select a valid language.",
    )
    """2-letter language code."""
    flags = models.PositiveSmallIntegerField(
        default=0,
        choices=[
            (0, _("Trigger on first message")),
            (1, _("Trigger on every message")),
            (2, _("Disabled")),
        ],
    )
    """Flags."""

    timezone = models.IntegerField(default=0)
    """Timezone."""
    unit_list = models.CharField(blank=True, max_length=2048)
    """List of Wialon units."""
    trigger = models.JSONField(blank=True, default=dict)
    """Trigger."""
    schedule = models.JSONField(blank=True, default=dict)
    """Schedule."""
    control_schedule = models.JSONField(blank=True, default=dict)
    """Control schedule."""
    actions = models.JSONField(blank=True, default=dict)
    """Actions."""
    text = models.CharField(max_length=1024, blank=True)
    """Text."""
    enabled = models.BooleanField(default=True)
    """Whether the notification is enabled in Wialon."""

    wialon_id = models.PositiveBigIntegerField()
    """Wialon id."""
    customer = models.ForeignKey(
        "terminusgps_notifications.Customer",
        on_delete=models.CASCADE,
        related_name="notifications",
    )
    """Associated customer."""
    date_created = models.DateTimeField(auto_now_add=True)
    """Date created."""

    class Meta:
        verbose_name = _("wialon notification")
        verbose_name_plural = _("wialon notifications")

    def __str__(self) -> str:
        """Returns the notification name."""
        return str(self.name)

    def save(self, **kwargs) -> None:
        if update_fields := kwargs.get("update_fields"):
            print(f"{update_fields = }")
        return super().save(**kwargs)

    def get_absolute_url(self) -> str:
        return reverse(
            "terminusgps_notifications:detail notifications",
            kwargs={"notification_pk": self.pk},
        )

    def get_text(self) -> str:
        """Returns the notification text (txt)."""
        return urllib.parse.urlencode(
            {
                "unit_id": "%UNIT_ID%",
                "user_id": self.customer.user.pk,
                "message": self.message,
            }
        )

    def get_actions(self) -> list[dict[str, typing.Any]]:
        """Returns a list of notification actions (act)."""
        return [
            {
                "t": "push_messages",
                "p": {
                    "url": urllib.parse.urljoin(
                        "https://api.terminusgps.com/",
                        f"/v3/notify/{self.method}/",
                    ),
                    "get": 1,  # 1 - GET request, 2 - POST request
                },
            }
        ]

    @transaction.atomic
    def enable(self, session: WialonSession) -> None:
        """Enables the notification in Wialon."""
        try:
            if not self.enabled:
                session.wialon_api.resource_update_notification(
                    **{
                        "itemId": int(self.customer.resource_id),
                        "id": self.wialon_id,
                        "callMode": "enable",
                        "e": 1,
                    }
                )
                self.enabled = True
        except WialonAPIError as e:
            logger.critical(e)
            raise

    @transaction.atomic
    def disable(self, session: WialonSession) -> None:
        """Disables the notification in Wialon."""
        try:
            if self.enabled:
                session.wialon_api.resource_update_notification(
                    **{
                        "itemId": int(self.customer.resource_id),
                        "id": self.wialon_id,
                        "callMode": "enable",
                        "e": 0,
                    }
                )
                self.enabled = False
        except WialonAPIError as e:
            logger.critical(e)
            raise

    def get_wialon_parameters(self, call_mode: str) -> dict[str, typing.Any]:
        """Returns parameters for Wialon notification API calls."""
        allowed_call_modes: set[str] = {"create", "update", "delete"}
        if call_mode not in allowed_call_modes:
            raise ValueError(
                f"Invalid call_mode '{call_mode}'. Options are: {allowed_call_modes}"
            )
        if not self.customer.resource_id:
            raise ValueError(
                f"{self.customer} didn't have a Wialon resource id set."
            )
        return {
            "itemId": int(self.customer.resource_id),
            "id": 0 if call_mode == "create" else self.wialon_id,
            "callMode": call_mode,
            "n": self.name,
            "txt": self.text,
            "ta": self.activation_time if self.activation_time else 0,
            "td": self.deactivation_time if self.deactivation_time else 0,
            "ma": self.max_alarms,
            "mmtd": self.max_message_interval,
            "cdt": self.alarm_timeout,
            "mast": self.min_duration_alarm,
            "mpst": self.min_duration_prev,
            "cp": self.control_period,
            "fl": self.flags,
            "la": self.language,
            "tz": self.timezone,
            "un": ast.literal_eval(self.unit_list),
            "trg": self.trigger,
            "act": self.actions,
            "sch": self.schedule,
            "ctrl_sch": self.control_schedule,
        }

    @transaction.atomic
    def delete_in_wialon(
        self, session: WialonSession
    ) -> dict[str, typing.Any]:
        try:
            params = self.get_wialon_parameters(call_mode="delete")
            return session.wialon_api.resource_update_notification(**params)
        except WialonAPIError as e:
            logger.critical(e)
            raise

    @transaction.atomic
    def update_in_wialon(
        self, session: WialonSession
    ) -> dict[str, typing.Any]:
        """Updates the notification in Wialon."""
        try:
            params = self.get_wialon_parameters(call_mode="update")
            return session.wialon_api.resource_update_notification(**params)
        except WialonAPIError as e:
            logger.critical(e)
            raise

    @transaction.atomic
    def create_in_wialon(
        self, session: WialonSession
    ) -> dict[str, typing.Any]:
        """Creates the notification in Wialon."""
        try:
            params = self.get_wialon_parameters(call_mode="create")
            return session.wialon_api.resource_update_notification(**params)
        except WialonAPIError as e:
            logger.critical(e)
            raise



from django import forms

from terminusgps_notifications.models import WialonNotification


class WialonNotificationUpdateForm(forms.ModelForm):
    class Meta:
        model = WialonNotification
        fields = [
            "name",
            "message",
            "method",
            "activation_time",
            "deactivation_time",
            "max_alarms",
            "max_message_interval",
            "alarm_timeout",
            "control_period",
            "min_duration_alarm",
            "min_duration_prev",
            "language",
            "flags",
        ]
        help_texts = {
            "message": "Enter a message to send to the destination phone number. You may use the Wialon tags detailed below."
        }
        widgets = {
            # Required fields
            "name": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 user-invalid:bg-red-50 user-invalid:text-red-600",
                    "placeholder": "New Notification",
                    "enterkeyhint": "next",
                    "inputmode": "text",
                }
            ),
            "message": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "placeholder": "Hello! At %MSG_TIME%, your vehicle %UNIT% entered %ZONE%.",
                    "enterkeyhint": "next",
                    "inputmode": "text",
                }
            ),
            "method": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            # Advanced fields
            "activation_time": forms.widgets.DateTimeInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "deactivation_time": forms.widgets.DateTimeInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "max_alarms": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "max_message_interval": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "alarm_timeout": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "control_period": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "min_duration_alarm": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "min_duration_prev": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "language": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "flags": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
        }


class WialonNotificationCreationForm(forms.ModelForm):
    class Meta:
        model = WialonNotification
        fields = [
            "name",
            "message",
            "method",
            "timezone",
            "activation_time",
            "deactivation_time",
            "max_alarms",
            "max_message_interval",
            "alarm_timeout",
            "control_period",
            "min_duration_alarm",
            "min_duration_prev",
            "language",
            "flags",
            "customer",
            "trigger",
            "unit_list",
            "schedule",
            "control_schedule",
        ]
        help_texts = {
            "message": "Enter a message to send to the destination phone number. You may use the Wialon tags detailed below."
        }
        widgets = {
            # Required fields
            "name": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 user-invalid:bg-red-50 user-invalid:text-red-600",
                    "placeholder": "New Notification",
                    "enterkeyhint": "next",
                    "inputmode": "text",
                }
            ),
            "message": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "placeholder": "Hello! At %MSG_TIME%, your vehicle %UNIT% entered %ZONE%.",
                    "enterkeyhint": "next",
                    "inputmode": "text",
                }
            ),
            "method": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            # Advanced fields
            "activation_time": forms.widgets.DateTimeInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "deactivation_time": forms.widgets.DateTimeInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "max_alarms": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "max_message_interval": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "alarm_timeout": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "control_period": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "min_duration_alarm": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "min_duration_prev": forms.widgets.TextInput(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                    "size": 1,
                    "maxlength": 4,
                    "inputmode": "decimal",
                }
            ),
            "language": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            "flags": forms.widgets.Select(
                attrs={
                    "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
                }
            ),
            # Hidden fields
            "customer": forms.widgets.HiddenInput(),
            "trigger": forms.widgets.HiddenInput(),
            "unit_list": forms.widgets.HiddenInput(),
            "schedule": forms.widgets.HiddenInput(),
            "control_schedule": forms.widgets.HiddenInput(),
            "timezone": forms.widgets.HiddenInput(),
        }



from django import forms
from terminusgps_payments.models import AddressProfile, PaymentProfile


class PaymentProfileChoiceField(forms.ModelChoiceField):
    def label_from_instance(self, obj):
        if obj.credit_card is not None:
            cc = obj.credit_card
            return f"{cc.cardType} ending in {str(cc.cardNumber)[-4:]}"
        return str(obj)


class AddressProfileChoiceField(forms.ModelChoiceField):
    def label_from_instance(self, obj):
        if obj.address is not None:
            addr = obj.address
            return str(addr.address)
        return str(obj)


class CustomerSubscriptionCreationForm(forms.Form):
    payment_profile = PaymentProfileChoiceField(
        queryset=PaymentProfile.objects.all(),
        widget=forms.widgets.Select(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    address_profile = AddressProfileChoiceField(
        queryset=AddressProfile.objects.all(),
        widget=forms.widgets.Select(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    consent = forms.BooleanField(
        initial=False,
        widget=forms.widgets.CheckboxInput(
            attrs={"class": "accent-terminus-red-700"}
        ),
    )



from .customers import (
    AccountView,
    CustomerStatsView,
    CustomerSubscriptionCreateView,
    DashboardView,
    NotificationsView,
    SubscriptionView,
    WialonCallbackView,
    WialonLoginView,
)
from .main import HomeView
from .notifications import (
    WialonNotificationCreateView,
    WialonNotificationDeleteView,
    WialonNotificationDetailView,
    WialonNotificationListView,
    WialonNotificationTriggerParametersFormSuccessView,
    WialonNotificationTriggerParametersFormView,
    WialonNotificationTriggerSelectFormView,
    WialonNotificationUnitSelectFormView,
    WialonNotificationUpdateView,
)



import decimal
import json

from django import forms
from django.core.exceptions import ValidationError
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from .. import constants


class JSONDecimalEncoder(json.JSONEncoder):
    """Casts :py:type:`~decimal.Decimal` objects to :py:type:`float` before encoding."""

    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super().default(obj)


class TriggerForm(forms.Form):
    """Parent form for selecting/saving a notification trigger configuration."""

    un = forms.CharField(widget=forms.widgets.HiddenInput())
    """List of Wialon units."""
    p = forms.JSONField(encoder=JSONDecimalEncoder)
    """Trigger parameters."""
    t = forms.ChoiceField(
        choices=constants.WialonNotificationTriggerType.choices,
        help_text=_("Please select a trigger from the list."),
        initial=constants.WialonNotificationTriggerType.SENSOR,
        widget=forms.widgets.Select(
            {
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600",
                "hx-target": "#id_trigger_parameters",
                "hx-trigger": "load once, change",
                "hx-swap": "outerHTML",
                "hx-get": reverse_lazy(
                    "terminusgps_notifications:trigger parameters"
                ),
            }
        ),
    )
    """Trigger type."""


class TriggerParametersForm(forms.Form):
    """Base class for notification trigger parameter forms."""

    def as_json(self) -> str:
        return json.dumps(self.cleaned_data, cls=JSONDecimalEncoder)


class GeofenceTriggerParametersForm(TriggerParametersForm):
    lower_bound = forms.DecimalField(label="Lower bound", initial=-1)
    upper_bound = forms.DecimalField(label="Upper bound", initial=1)
    sensor_name_mask = forms.CharField(label="Sensor mask", initial="*")
    sensor_type = forms.ChoiceField(
        label="Sensor type",
        choices=constants.WialonUnitSensorType.choices,
        initial=constants.WialonUnitSensorType.ANY,
        required=False,
    )
    geozone_ids = forms.CharField()
    min_speed = forms.IntegerField(
        label="Min speed", min_value=0, max_value=256, initial=0
    )
    max_speed = forms.IntegerField(
        label="Max speed", min_value=0, max_value=256, initial=256
    )
    prev_msg_diff = forms.TypedChoiceField(
        choices=[
            (0, _("Relative to absolute value")),
            (1, _("Relative to previous value")),
        ],
        coerce=int,
    )
    merge = forms.TypedChoiceField(
        label="Similar sensors",
        initial=0,
        choices=[(0, _("Calculate separately")), (1, _("Sum up values"))],
        coerce=int,
    )
    reversed = forms.TypedChoiceField(
        label="Boundaries",
        initial=0,
        choices=[
            (0, _("In the specified range")),
            (1, _("Out of the specified range")),
        ],
        coerce=int,
    )
    type = forms.TypedChoiceField(
        label="Trigger when",
        initial=0,
        choices=[
            (0, _("Entering geofence(s)")),
            (1, _("Exiting geofence(s)")),
        ],
        coerce=int,
    )
    include_lbs = forms.BooleanField(
        initial=False, label="Process LBS messages", required=False
    )
    lo = forms.ChoiceField(
        choices=[("", "None"), ("AND", _("AND")), ("OR", _("OR"))],
        initial="",
        label="Logical operator",
        required=False,
    )


class AddressTriggerParametersForm(TriggerParametersForm):
    city = forms.CharField()
    country = forms.CharField()
    house = forms.CharField()
    include_lbs = forms.ChoiceField()
    lower_bound = forms.DecimalField()
    max_speed = forms.IntegerField()
    merge = forms.ChoiceField()
    min_speed = forms.IntegerField()
    prev_msg_diff = forms.TypedChoiceField(
        initial=0,
        choices=[
            (0, _("Relative to absolute value")),
            (1, _("Relative to previous value")),
        ],
        coerce=int,
    )
    radius = forms.IntegerField()
    region = forms.CharField()
    reversed = forms.TypedChoiceField(
        label="Boundaries",
        initial=0,
        choices=[
            (0, _("In the specified range")),
            (1, _("Out of the specified range")),
        ],
        coerce=int,
    )
    sensor_name_mask = forms.CharField(label="Sensor mask", initial="*")
    sensor_type = forms.ChoiceField(
        choices=constants.WialonUnitSensorType.choices,
        initial=constants.WialonUnitSensorType.ANY,
    )
    street = forms.CharField()
    type = forms.ChoiceField()
    upper_bound = forms.DecimalField()


class SpeedTriggerParametersForm(TriggerParametersForm):
    lower_bound = forms.DecimalField()
    max_speed = forms.IntegerField()
    merge = forms.ChoiceField()
    min_speed = forms.IntegerField()
    prev_msg_diff = forms.TypedChoiceField(
        choices=[
            (0, _("Relative to absolute value")),
            (1, _("Relative to previous value")),
        ],
        coerce=int,
    )
    reversed = forms.TypedChoiceField(
        label="Boundaries",
        initial=0,
        choices=[
            (0, _("In the specified range")),
            (1, _("Out of the specified range")),
        ],
        coerce=int,
    )
    sensor_name_mask = forms.CharField()
    sensor_type = forms.ChoiceField(
        choices=constants.WialonUnitSensorType.choices,
        initial=constants.WialonUnitSensorType.ANY,
    )
    upper_bound = forms.IntegerField(initial=1)


class AlarmTriggerParametersForm(TriggerParametersForm):
    pass


class DigitalTriggerParametersForm(TriggerParametersForm):
    input_index = forms.IntegerField(min_value=1, max_value=32)
    type = forms.TypedChoiceField(
        choices=[
            (0, _("Check for activation")),
            (1, _("Check for deactivation")),
        ],
        coerce=int,
    )


class ParameterTriggerParametersForm(TriggerParametersForm):
    kind = forms.TypedChoiceField(
        choices=[
            (0, _("Value range")),
            (1, _("Text mask")),
            (2, _("Parameter availability")),
            (3, _("Parameter lack")),
        ],
        coerce=int,
    )
    lower_bound = forms.DecimalField()
    param = forms.CharField()
    text_mask = forms.CharField()
    type = forms.TypedChoiceField(
        choices=[
            (0, _("In the specified range")),
            (1, _("Out of the specified range")),
        ],
        coerce=int,
    )
    upper_bound = forms.DecimalField()


class SensorValueTriggerParametersForm(TriggerParametersForm):
    lower_bound = forms.DecimalField(
        label="Lower bound",
        initial=-1,
        widget=forms.widgets.TextInput(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    upper_bound = forms.DecimalField(
        label="Upper bound",
        initial=1,
        widget=forms.widgets.TextInput(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    merge = forms.TypedChoiceField(
        label="Similar sensors",
        initial=0,
        choices=[(0, _("Calculate separately")), (1, _("Sum up values"))],
        coerce=int,
        widget=forms.widgets.Select(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    prev_msg_diff = forms.TypedChoiceField(
        label="Boundaries",
        choices=[
            (0, _("Relative to absolute value")),
            (1, _("Relative to previous value")),
        ],
        coerce=int,
        widget=forms.widgets.Select(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    sensor_name_mask = forms.CharField(
        label="Sensor mask",
        initial="*",
        widget=forms.widgets.TextInput(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    sensor_type = forms.ChoiceField(
        choices=constants.WialonUnitSensorType.choices,
        initial="",
        label="Sensor type",
        required=False,
        widget=forms.widgets.Select(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )
    type = forms.TypedChoiceField(
        label="Trigger when",
        initial=0,
        choices=[
            (0, _("Sensor value in-range")),
            (1, _("Sensor value out-of-range")),
        ],
        coerce=int,
        widget=forms.widgets.Select(
            attrs={
                "class": "peer p-2 rounded border border-current bg-gray-50 dark:bg-gray-600 invalid:bg-red-50 invalid:text-red-600"
            }
        ),
    )

    def clean(self) -> None:
        cleaned_data = super().clean()
        upper_bound = cleaned_data.get("upper_bound")
        lower_bound = cleaned_data.get("lower_bound")
        if upper_bound is not None and lower_bound is not None:
            if lower_bound > upper_bound:
                self.add_error(
                    "lower_bound",
                    ValidationError(
                        _(
                            "Lower bound cannot be greater than upper bound (%(lower_bound)s > %(upper_bound)s)."
                        ),
                        params={
                            "lower_bound": float(lower_bound),
                            "upper_bound": float(upper_bound),
                        },
                    ),
                )
            elif upper_bound < lower_bound:
                self.add_error(
                    "upper_bound",
                    ValidationError(
                        _(
                            "Upper bound cannot be less than lower bound (%(upper_bound)s < %(lower_bound)s)."
                        ),
                        params={
                            "lower_bound": float(lower_bound),
                            "upper_bound": float(upper_bound),
                        },
                    ),
                )


class OutageTriggerParametersForm(TriggerParametersForm):
    check_restore = forms.TypedChoiceField(
        choices=[
            (0, _("Connection lost")),
            (1, _("Connection lost and restored")),
            (2, _("Connection restored")),
        ],
        coerce=int,
    )
    geozones_list = forms.CharField()
    geozones_type = forms.TypedChoiceField(
        choices=[(0, _("Out of geofence")), (1, _("In geofence"))], coerce=int
    )
    include_lbs = forms.BooleanField()
    time = forms.IntegerField()
    type = forms.TypedChoiceField(
        choices=[(0, _("Coordinates loss")), (1, _("Connection loss"))],
        coerce=int,
    )


class InterpositionTriggerParametersForm(TriggerParametersForm):
    include_lbs = forms.BooleanField()
    lo = forms.ChoiceField(
        choices=[("AND", _("Logical AND")), ("OR", _("Logical OR"))]
    )
    lower_bound = forms.DecimalField()
    max_speed = forms.IntegerField()
    merge = forms.TypedChoiceField(
        choices=[(0, _("Calculate separately")), (1, _("Sum up values"))],
        coerce=int,
    )
    min_speed = forms.IntegerField()
    prev_msg_diff = forms.TypedChoiceField(
        choices=[
            (0, _("Relative to absolute value")),
            (1, _("Relative to previous value")),
        ],
        coerce=int,
    )
    radius = forms.IntegerField()
    reversed = forms.TypedChoiceField(
        label="Boundaries",
        initial=0,
        choices=[
            (0, _("In the specified range")),
            (1, _("Out of the specified range")),
        ],
        coerce=int,
    )
    sensor_name_mask = forms.CharField(label="Sensor mask", initial="*")
    sensor_type = forms.ChoiceField(
        choices=constants.WialonUnitSensorType.choices,
        initial=constants.WialonUnitSensorType.ANY,
    )
    type = forms.TypedChoiceField(
        choices=[
            (0, _("Control approaching to units")),
            (1, _("Control moving away from units")),
        ],
        coerce=int,
    )
    unit_guids = forms.CharField()
    upper_bound = forms.DecimalField()


class ExcessTriggerParametersForm(TriggerParametersForm):
    flags = forms.TypedChoiceField(
        choices=[(1, _("Data messages")), (2, _("SMS messages"))], coerce=int
    )
    msgs_limit = forms.IntegerField()
    time_offset = forms.IntegerField()


class RouteTriggerParametersForm(TriggerParametersForm):
    mask = forms.CharField()
    round_mask = forms.CharField()
    schedule_mask = forms.CharField()
    types = forms.CharField()


class DriverTriggerParametersForm(TriggerParametersForm):
    driver_code_mask = forms.CharField()
    flags = forms.TypedChoiceField(
        choices=[(1, _("Driver assignment")), (2, _("Driver separation"))],
        coerce=int,
    )


class TrailerTriggerParametersForm(TriggerParametersForm):
    driver_code_mask = forms.CharField()
    flags = forms.TypedChoiceField(
        choices=[(1, _("Trailer assignment")), (2, _("Trailer separation"))],
        coerce=int,
    )


class MaintenanceTriggerParametersForm(TriggerParametersForm):
    days = forms.IntegerField()
    engine_hours = forms.IntegerField()
    flags = forms.TypedChoiceField(
        choices=[
            (0, _("Control all service intervals")),
            (1, _("Mileage interval")),
            (2, _("Engine hours interval")),
            (4, _("Days interval")),
        ],
        coerce=int,
    )
    mask = forms.CharField()
    mileage = forms.IntegerField()
    val = forms.TypedChoiceField(
        choices=[
            (1, _("Notify when service term approaches")),
            (-1, _("Notify when service term is expired")),
        ],
        coerce=int,
    )


class FuelTriggerParametersForm(TriggerParametersForm):
    sensor_name_mask = forms.CharField(label="Sensor mask", initial="*")
    geozones_type = forms.TypedChoiceField(
        choices=[
            (0, _("Disabled/outside geofence")),
            (1, _("Inside geofence")),
        ],
        coerce=int,
    )
    geozones_list = forms.CharField()
    realtime_only = forms.BooleanField()


class FuelDrainTriggerParametersForm(TriggerParametersForm):
    sensor_name_mask = forms.CharField(label="Sensor mask", initial="*")
    geozones_type = forms.TypedChoiceField(
        choices=[
            (0, _("Disabled/outside geofence")),
            (1, _("Inside geofence")),
        ],
        coerce=int,
    )
    geozones_list = forms.CharField()
    realtime_only = forms.BooleanField()


class HealthTriggerParametersForm(TriggerParametersForm):
    healthy = forms.TypedChoiceField(
        choices=[
            (0, _("Don't trigger when the device is healthy")),
            (1, _("Trigger when the device is healthy")),
        ],
        coerce=int,
    )
    unhealthy = forms.TypedChoiceField(
        choices=[
            (0, _("Don't trigger when the device is unhealthy")),
            (1, _("Trigger when the device is unhealthy")),
        ],
        coerce=int,
    )
    needAttention = forms.TypedChoiceField(
        choices=[
            (0, _("Don't trigger when the device needs attention")),
            (1, _("Trigger when the device needs attention")),
        ],
        coerce=int,
    )
    triggerForEachIncident = forms.TypedChoiceField(
        choices=[
            (0, _("Don't trigger for each incident")),
            (1, _("Trigger for each incident")),
        ],
        coerce=int,
    )


TRIGGER_FORM_MAP = {
    constants.WialonNotificationTriggerType.GEOFENCE: GeofenceTriggerParametersForm,
    constants.WialonNotificationTriggerType.ADDRESS: AddressTriggerParametersForm,
    constants.WialonNotificationTriggerType.SPEED: SpeedTriggerParametersForm,
    constants.WialonNotificationTriggerType.ALARM: AlarmTriggerParametersForm,
    constants.WialonNotificationTriggerType.DIGITAL: DigitalTriggerParametersForm,
    constants.WialonNotificationTriggerType.PARAMETER: ParameterTriggerParametersForm,
    constants.WialonNotificationTriggerType.SENSOR: SensorValueTriggerParametersForm,
    constants.WialonNotificationTriggerType.OUTAGE: OutageTriggerParametersForm,
    constants.WialonNotificationTriggerType.INTERPOSITION: InterpositionTriggerParametersForm,
    constants.WialonNotificationTriggerType.EXCESS: ExcessTriggerParametersForm,
    constants.WialonNotificationTriggerType.ROUTE: RouteTriggerParametersForm,
    constants.WialonNotificationTriggerType.DRIVER: DriverTriggerParametersForm,
    constants.WialonNotificationTriggerType.TRAILER: TrailerTriggerParametersForm,
    constants.WialonNotificationTriggerType.MAINTENANCE: MaintenanceTriggerParametersForm,
    constants.WialonNotificationTriggerType.FUEL: FuelTriggerParametersForm,
    constants.WialonNotificationTriggerType.FUEL_DRAIN: FuelDrainTriggerParametersForm,
    constants.WialonNotificationTriggerType.HEALTH: HealthTriggerParametersForm,
}



from django.apps import AppConfig


class TerminusgpsNotificationsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "terminusgps_notifications"
    verbose_name = "Terminus GPS Notifications"



from django.views.generic import TemplateView
from terminusgps.mixins import HtmxTemplateResponseMixin


class HomeView(HtmxTemplateResponseMixin, TemplateView):
    content_type = "text/html"
    extra_context = {
        "title": "Terminus GPS Notifications",
        "subtitle": "We know where ours are... do you?",
    }
    http_method_names = ["get"]
    partial_template_name = "terminusgps_notifications/partials/_home.html"
    template_name = "terminusgps_notifications/home.html"
